// 
// Module: CalResponse
// Author: Hassan Siddiqui
//
// Description: 
// Provides a clean interface to CAL response data.
// 
// Comments: 
//  
// Modifications: See ChangeLog
// 
// $Id: CalResponse.cc,v 1.1 2000/09/21 15:34:34 hsiddiqu Exp $

/* General */
#include <string>
#include <iostream>
#include <cmath>
#include <vector>

/* SAS-specific */
#include "Dal.h"
#include "CalServer.h"
#include "Redistribution.h"
#include "Tracer.h"
#include "NumericalRange.h"
#include "IntervalRebinner.h"

// Local classes
#include "CalResponse.h"

/*****************************
 *
 * class CalResponse
 *
 * Interface to CAL response data.
 *
 *****************************/

/** Constructors */
CalResponse::CalResponse()
{
  PROC;
  ENDPROC;
}

/** Initialisation Methods */

/* Core initialisation */
void CalResponse::create ( unsigned lowerPI,
			   unsigned nOutputChannels,
			   unsigned outputChannelWidth)
{
  _lastChannel = nOutputChannels-1;

  errstr 
	 << "Creating CalResponse object for "
	 << nOutputChannels
	 << " output Channels, starting from PI"
	 << lowerPI
	 << " with a bin width of "
	 << outputChannelWidth
	 << " PI channels. "
	 << Sas::message(Msg::App, Msg::Verbose);

  _rowLow  = 0;

  _nrows   = _ebins->bins();
  _rowHigh = _nrows - 1;
  
  _energyLow  = _ebins->lower();
  _energyHigh = _ebins->upper();

  errstr 
    << " energyLow " << _energyLow 
    << " energyHigh " << _energyHigh
    << Sas::message(Msg::App,Msg::Verbose);

  /* Get the energy bounds for the PI channels from the CAL */
  const BinnedAxis * standardBounds = & _rmf->binnedPIeBounds();

  /* The standard boundaries are mapped for a given P.I. binning,
     e.g. 5 PI channels per bin for the MOS. This is a weak link and
     is difficult to test for properly. Here we bin up the standard
     boundaries if the number of channels required is an integer divisor
     of the number of standard bins and if not we map a uniform axis */

  //errstr<<"Standard[1]: "<<standardBounds->bin(1).lo()
//	      << Sas::message(Msg::App,Msg::Verbose);

  unsigned nstandard = standardBounds->bins();

  unsigned rebinFact = UNT( nstandard / nOutputChannels );

  errstr<<"Rebinfactor: "<<rebinFact
        <<"CCFlength: "<<nstandard
        <<"nOutputChannels: "<<nOutputChannels
        <<" diff: "<<abs(double(nstandard / nOutputChannels) - rebinFact)
        <<" thediv: "<<double(nstandard / nOutputChannels)
        << Sas::message(Msg::App,Msg::Verbose);

  //IntervalVector<double> iv;
  //unsigned s=0;

  if ( abs(double(nstandard) / double(nOutputChannels) - rebinFact) <= FLT64_MIN ) 
  {
    /* Set the actual bounds axis object */
     _eboundsAxis = new NonUnifBinnedAxis(
                    IntervalRebinner::rebin(standardBounds, rebinFact, 0) );
  }
  else {

     errstr<<"Can't rebin channel boundaries from the CCF"
           << " - using regular bins "
           << Sas::warning("invalidCCFChannels");

     _eboundsAxis = new UniformilyBinnedAxis(double(lowerPI),
                    double(lowerPI + nOutputChannels * outputChannelWidth),
                            nOutputChannels );

  }

}

/* Initialise using CAL ebins */
void CalResponse::createWithCalEbins ( unsigned lowerPI,
				       unsigned nOutputChannels,
				       unsigned outputChannelWidth )
{
  _ebins = &_rmf->eBins();
  create(lowerPI, nOutputChannels, outputChannelWidth);
}

/* Initialise using user-defined ebins */
void CalResponse::createWithUserEbins ( unsigned lowerPI,
					unsigned nOutputChannels,
					unsigned outputChannelWidth,
					const BinnedAxis  * ebins )
{
  _ebins = ebins;
  create(lowerPI, nOutputChannels, outputChannelWidth);
}

/** Misc Methods */
void CalResponse::check()
{

      PROC;

      /* Get actual number of channels available from data if this differs
       * to requested range, issue a message.*/

      unsigned numberofactualchannels = _rmf->eBounds().bins();

      if ( numberOfChannels() > numberofactualchannels)
	  {
	    errstr
	      << " Number of channels requested [ "
	      << numberOfChannels()
	      << " ] is greater than number of channels available from CAL ["
	      << numberofactualchannels
	      <<"]. Will pad out data to size requested. "
	      << Sas::message(Msg::App,Msg::Sparse);
	  }
      else if (numberOfChannels() < numberofactualchannels)
	{
	  errstr 
	      << " Number of channels requested [ "
	      << numberOfChannels()
	      << " ] is less than number of channels available from CAL ["
	      << numberofactualchannels
	      <<"]. Will pad out data to size requested. "
	      << Sas::message(Msg::App,Msg::Sparse);
	}

      ENDPROC;
}

/* Pad vector with zeros, or truncate vector, to a given size. */
const vector<double> CalResponse::resizeVector( const vector<double> & in, 
						unsigned nelements) const
{
  PROC;

  vector<double> out;

      if (nelements == in.size())
	{
	  ALTRETURNV(in);
	}

      errstr 
	<< "resizeVector: size of vector ["<<in.size()
	<<"] differs from number of elements in matrix [" << nelements << "]"
	<< Sas::message(Msg::App, Msg::Sparse);


      if (nelements < in.size())
	{
	  errstr << " Size of vector is " << in.size()
		 << " truncating to " << nelements
		 << Sas::message(Msg::App, Msg::Verbose);

	  /* Truncate vector to nelements */
	  for (unsigned i = 0; i < nelements; ++i)
	    { 
	      out.push_back(in[i]);
	    }
	  ALTRETURNV(out);
	}

      /* Pad out vector to nelements */
      for (unsigned i = 0; i < in.size(); ++i)
	{
	  out.push_back(in[i]);
	}
      for (unsigned i = in.size(); i < nelements; ++i)
	{
	  out.push_back(0.);
	}
      RETURN out;
}

/** Destructor */
CalResponse::~CalResponse()
{
  PROC;
  delete _eboundsAxis; _eboundsAxis = 0;
  ENDPROC;
}


/** Channel range accessors */

const unsigned CalResponse::firstChannel()     const 
    { 
      PROC; 
      RETURN 0; 
    }

const unsigned CalResponse::lastChannel()      const 
    { 
      PROC; RETURN _lastChannel;
    }

  /** Response Accessors */

  unsigned                 CalResponse::ebinFirstIndex()   const 
    {PROC; RETURN _rowLow;}

  unsigned                 CalResponse::numberOfEbins()    const 
    { PROC; RETURN _nrows; }

  const Bin CalResponse::ebins(unsigned row)    const 
    { 
      PROC; 

      // assert (row >= _rowLow && row <= _rowHigh);

      RETURN  _ebins->bin(row);
    }

  const vector<double>  CalResponse::responseVector ( unsigned row ) const
    { 
      PROC;

      // assert (row >= _rowLow && row <= _rowHigh);

      Bin bin =_ebins->bin(row);
      double energy = 0.5 * (bin.lo() + bin.hi());

      errstr 
	<< " Returning a response for energy "
	<<  energy
	<< " eV. "
	<< Sas::message(Msg::App, Msg::Noisy);

     /* Currently the instrument teams blank out everything below
        30 eV - so we zero the low-energy response here */

      vector<double> result;

      if (energy < 30.0) {
         for (unsigned i=0; i<_eboundsAxis->bins(); i++) {
             result.push_back(0.0);
         }
      }
      else {
         result = _rmf->spectrum( energy, *_eboundsAxis, _eventGrade);
      }

   RETURN resizeVector( result, 
			numberOfChannels()); 

    }
  
  const vector<double> CalResponse::responseVector ( double energy ) const
    { 
      PROC;
      RETURN resizeVector( _rmf->spectrum(energy, *_eboundsAxis, _eventGrade), 
			   numberOfChannels() ); 
    }
  

  /* Ebounds Accessor 
   * Note the size and range corresponds to that specified for the channel,
   * so no accessors are needed to cover this. */

  const BinnedAxis * CalResponse::ebounds()  const 
    { PROC; RETURN _eboundsAxis; }
