// Module: LowMemFixedMatrixColumn
// Author: Hassan Siddiqui
//
// Description:
// Fixed length vector column version of MatrixColumn for DAL LowMem implementation.
// 
// Comments: 
//  
// Modifications: See ChangeLog
// 
// $Id: LowMemFixedMatrixColumn.cc,v 1.3 2000/03/28 12:57:23 hsiddiqu Exp $

/* General */

/* SAS-specific */
#include "Dal.h"
#include "errstr.h"
#include "Tracer.h"

/* Local */
#include "LowMemFixedMatrixColumn.h"

/* Constructor */
LowMemFixedMatrixColumn::LowMemFixedMatrixColumn( Table * table,
				      const vector<unsigned long>& sizeVect):
  _matrixCol(0),
  _matrixColData(0),
  _matrixData(0),
  _row(0),
  _largestelement(0)
{
  PROC;

  errstr << " creating LowMemFixedMatrixColumn object ... "
	 << Sas::message(Msg::App, Msg::Verbose);

  /* Check that sizeVect is scalar */
  if (sizeVect.size() != 1)
    {
      errstr << " sizeVect has " 
	     << sizeVect.size() << " dimensions."
	     << " expected 1. "
	     << Sas::error("matrixDimensionsInvalid");
      ALTRETURN;
    }

  /* Create new Column */
  _matrixCol = table->addColumn("MATRIX", 
				Column::Real32, 
				"No comment", 
				"", 
				sizeVect);

  _rowelements = sizeVect[0];

  errstr << " matrixColData " << _matrixColData
	 << Sas::message(Msg::App, Msg::Verbose);

  errstr << " LowMemFixedMatrixColumn object created. "
	 << Sas::message(Msg::App, Msg::Verbose);

  ENDPROC;
}


/* Destructor */
LowMemFixedMatrixColumn::~LowMemFixedMatrixColumn()
{
  PROC;
  
  errstr << " Destroying LowMemFixedMatrixColumn object ... "
	 << Sas::message(Msg::App, Msg::Verbose);
 
  if (_matrixColData)
    {
      errstr << " matrixColData object for LowMemFixedMatrixColumn is still set before destructor was accessed. This should not occur, please raise an SPR."
	     << Sas::error("invalidMatrixColData");
      ALTRETURN;
    }
  ENDPROC;
}

/* Set state to a particular row, and the size of that row */
void LowMemFixedMatrixColumn::row( unsigned long row,
			     unsigned int /* elements NOT USED */ )
{
  PROC;
  errstr << "LowMemFixedMatrixColumn::row() called."
	 << Sas::message(Msg::App, Msg::Verbose);
  
  if (_matrixColData)
    {
      errstr << " matrixColData object for LowMemFixedMatrixColumn is still set ["
	     << _matrixColData
	     << "] before row() was accessed. This should not occur, please raise an SPR."
	     << Sas::error("invalidMatrixData");
      ALTRETURN;
    }
  
  _matrixColData = _matrixCol->data(row,1);
  _matrixData    = _matrixColData  ->real32Data();

  _row = row;
  
  ENDPROC;
}

/* Write to a given position in column */
void LowMemFixedMatrixColumn::writeTo( unsigned int element,
				 double value)
{
  PROC;

  errstr << "LowMemFixedMatrixColumn::writeTo() called."
	 << Sas::message(Msg::App, Msg::Verbose);
  
  if ( element > _rowelements)
    {
      errstr << " element specified [" << element
	     << "] is out of bounds for matrix column [0:" << _rowelements-1
	     << Sas::error("elementOutOfBounds");
      ALTRETURN;
    }

  if (!_matrixColData)
    {
      errstr << " matrixColData object for LowMemFixedMatrixColumn is not set in writeTo() method. This should not occur, please raise an SPR."
	     << Sas::error("invalidMatrixColData");
      ALTRETURN;
    }

  _matrixData[_matrixColData->position( _row, element)] = value;

  errstr << "matrix ["
	 << _row
	 << ","
	 << element
	 << "] set to "
	 << _matrixData[_matrixColData->position(_row, element)]
	 << "( expected " << value << " )"
	 << Sas::message(Msg::App, Msg::Verbose);

  /* Set _largestelement if appropriate */
  if (element > _largestelement)
    {
      _largestelement = element;
    }

  ENDPROC;
}

unsigned int LowMemFixedMatrixColumn::rowElements() const
{
  PROC;
  errstr << "LowMemFixedMatrixColumn::rowElements() called."
	 << Sas::message(Msg::App, Msg::Verbose);

  RETURN _rowelements;
}


void LowMemFixedMatrixColumn::closeRow()
{
  PROC;

  errstr << "LowMemFixedMatrixColumn::closeRow() called."
	 << Sas::message(Msg::App, Msg::Verbose);

  if (!_matrixColData)
    {
      errstr << " matrixColData object for LowMemFixedMatrixColumn is not set in closeRow() method. This should not occur, please raise an SPR."
	     << Sas::error("invalidMatrixColData");
      ALTRETURN;
    }

  if (_largestelement<_rowelements-1)
    {
      /* Pad remainder of row with zeroes */
      errstr << " Padding row " << _row
	     << " elements " << _largestelement+1
	     << " to " << _rowelements-1
	     << " with zeroes..."
	     << Sas::message(Msg::App, Msg::Verbose);

      for (unsigned int element = _largestelement+1; 
	   element <_rowelements;
	   ++element)
	{
	  _matrixData[_matrixColData->position( _row, element)] = 0;
	}
    }

  delete _matrixColData; 
  _matrixColData = 0;

  errstr << " matrixColData " << _matrixColData
	 << Sas::message(Msg::App, Msg::Verbose);

  ENDPROC;
}
