// Module: ResponseRow
// Author: Hassan Siddiqui
//
// Description: Classes which allow the response matrix to be groups.
// 
// Comments: 
//  
// Modifications: See ChangeLog
// 
// $Id: ResponseRow.cc,v 1.2 2000/03/07 13:49:44 hsiddiqu Exp $

/* General */
#include <iostream>
//#include "list.h"

/* SAS-specific classes */
#include "Dal.h"
#include "NumericalRange.h"
#include "Tracer.h"
#include "errstr.h"

/* Local */
#include "ResponseRow.h"

/******************************************************************
 *								   
 * class ResponseRow
 *
 * Contains statistics (# groups, # elements > threshold ) 
 * and grouped data for a row within the response matrix 
 * Later, this might be split into two classes: none which generates the statistics
 * and the other which generates statistics as well as the grouped data structure.
 *
 ******************************************************************/

/* Use constructor using as input the row of interest 
 * of the raw rmf as input to generate a grouped response.
 */
ResponseRow::ResponseRow( const vector<double>& response, double threshold) : 
  _threshold(threshold), 
  _ngroups(0), 
  _nresponse_elements(0)
{
  
  PROC;

 /* Find sum of elements above threshold for subsequent normalisation */
  double sum = 0.0;

  for (unsigned int el=0; el < response.size(); el++) 
  {
     if (response[el] >= _threshold)
     {
        sum += response[el];
     }
  }

  if (sum <= FLT64_MIN) { sum = 1.0; }

  // Set up a 'perpetual' loop to accumulate groups.
  // This is broken when all the elements in the row of the response are read. 
  
  unsigned long element = 0;
  
  while (true)
    {
      
      // Skip non-valid elements 
      
      while ( (response[element] <= _threshold) && (element < response.size()) )
	{
	  ++element;
      if (element >= response.size()) break;
	}
	  
	  
      // Break from function if element number exceeds bounds for response array
      if (element >= response.size()) break;

      
      // Set up a new element grouping object
      
      ResponseGroup response_group(element);
      
      
      // Populate object with contiguous response element values above threshold.

      while ( (response[element] >= _threshold) && (element < response.size()) )
	{
	      
	  // Add response element to group
	  
	  response_group.push_back(ResponseElement(element, response[element]/sum));

	  ++_nresponse_elements;

	  ++element;
	}
	  
      // Add current group to list of groups 

      push_back(response_group);
      _ngroups++;

    }

  // Check number of elements in row
  int nelements = 0;
  for ( ResponseRow::iterator group = begin(); group != end(); ++group)
    {
      nelements += (*group).size();
    }

  // assert (nelements == _nresponse_elements);

//    errstr << " ResponseRow: found " << _nresponse_elements
//  	 << Sas::message(Msg::App, Msg::Verbose);
  
  ENDPROC;
}

const vector<double> ResponseRow::expanded() const
{
  int position = 0;
  
  vector<double> result;
  
  for ( ResponseRow::const_iterator group = begin(); 
	group != end(); 
	++group)
    {
      /* Pad out output with zeroes */
      for (unsigned long el = 0; el < (*group).startofgroup(); ++el)
	{
	  result.push_back(0.);
	  position++;
	}
      
      for (ResponseGroup::const_iterator element = (*group).begin(); 
	   element != (*group).end(); 
	   ++element)
	{
	  result.push_back((*element).value());
	  position++;
	}
    }
      
  RETURN result;
}


void ResponseRow::writeTo(unsigned long row,
	     MatrixColumn * matrixColumn, 
	     ColumnData* n_grpColData,
	     ColumnData* f_chanColData,
	     ColumnData* n_chanColData, 
             unsigned tlmin) const
{
  PROC;
  
  int16*  n_grpData   = n_grpColData   ->int16Data();

  int16*  f_chanData  = f_chanColData  ->int16Data();

  int16*  n_chanData  = n_chanColData  ->int16Data();

  unsigned int group_index = 0;
      
  unsigned int element_index = 0;


  // Write out number of groups in list.

  n_grpData[row] = size();
	  
  /* If there are no groups for that row (ie no data)
   * fill columns with zeroes and return from function
   */

  if (size() == 0)
    {
      errstr << "No groups found for row " << row
	     << " Will fill F_CHAN, N_CHAN and MATRIX columns with zeroes "
	     << Sas::message(Msg::App, Msg::Sparse);

      for (unsigned int i = element_index; i < f_chanColData->elements(); ++i)
	{
	  f_chanData[f_chanColData->position( row, i)] = 0;
	}

      for (unsigned int i = element_index; i < n_chanColData->elements(); ++i)
	{
	  n_chanData[n_chanColData->position( row, i)] = 0;
	}

      /* The matrix column is a special case. The FixedMatrixColumn
	 representation has all elements 
	 in row initialised to zero already.
	 In the VariableLengthColumn representation, the row
	 is initialised when the row size is specified.
	 Set this here to 1. 
      */

      matrixColumn->row(row, 1);
      
//        errstr << "Number of elements in matrix " 
//  	     << matrixColumn->rowElements(row)
//  	     << Sas::message(Msg::App, Msg::Sparse);

//        for (unsigned int i = element_index; i < matrixColumn->rowElements(row); ++i)
//  	{
//  	  matrixColumn->writeTo( row, element_index, 0 );
//  	}
      
      matrixColumn->closeRow();

      ALTRETURN;
    }

  /* Inform matrixColumn of the true number of elements in row 
     (Ignored in the case of FixedMatrixColumn representations)
   */
  matrixColumn->row(row, _nresponse_elements);

  /* Zero all the F_CHAN and N_CHAN elements for this row */
    for (unsigned int i = element_index; i < f_chanColData->elements(); ++i)
	{
	  f_chanData[f_chanColData->position( row, i)] = 0;
	}

    for (unsigned int i = element_index; i < n_chanColData->elements(); ++i)
	{
	  n_chanData[n_chanColData->position( row, i)] = 0;
	}

  /* Loop through groups. */
  for ( ResponseRow::const_iterator group = begin(); 
	group != end(); 
	++group)
    {
     /* Set the F_CHAN (first channel) column */
	  
     // Channels can go from 0-(NCHANNELS-1) //
      //f_chanData[f_chanColData->position( row, group_index)] = 
	//(*(*group).begin()).channel() + 1; // Channels run from 1 - NCHANNELS
	  
     /* Set the F_CHAN and N_CHAN values for the defined groups */
      f_chanData[f_chanColData->position( row, group_index)] = 
	(*(*group).begin()).channel() + tlmin;
	  
      /* Write out number of channels in group. */
      n_chanData[n_chanColData->position( row, group_index)] = (*group).size();
	  

      /* Loop through elements in group. */
      for (ResponseGroup::const_iterator element = (*group).begin(); 
	   element != (*group).end(); 
	   ++element)
	{
	      
	  /* Write out element to matrix. */

	    //	    assert(element_index < matrixColumn->rowElements(row));
	    // assert ((*element).value() >= 0);

	    matrixColumn->writeTo(element_index, (*element).value());
	
	    /* Move onto index to next element in output matrix. */
	    ++element_index;
	    
	  }

	++group_index;
	    
    }

  // Fill remaining elements with zeroes

//    errstr 
//      << " Filling elements " << element_index
//      << " to " << matrixColumn->rowElements(row)
//      << " with zeroes "
//      << Sas::message(Msg::App, Msg::Verbose);

//    for (unsigned int i = element_index; i < matrixColumn->rowElements(row); ++i)
//      {
//        matrixColumn->writeTo(row, element_index, 0.);
//      }

  matrixColumn->closeRow();
  
  ENDPROC;
}
	     

// Accessors

int ResponseRow::ngroups() const
{ 
  PROC;
  //return _ngroups; 
  RETURN size(); 
}

int ResponseRow::nresponse_elements() const  
{
  PROC;
  RETURN _nresponse_elements; 
}

double ResponseRow::threshold() const {PROC RETURN _threshold; }




